Debugger = {
    thread = nil,
    debug_threads = false,
}

local saved_values = {}

local info = {
    traceback = {},
    locals = {}
}

local function gatherDebugInfo()
    
    local traceback = info.traceback
    local locals = info.locals
    
    local level = 1
    while true do
        local info = debug.getinfo(Debugger.thread.co, level + 2, "Sl")
        if not info then break end
        if info.what == "C" then   -- is a C function?
            table.insert(traceback, (level + 2) .. " C function")
            locals[level] = {}
            
        else   -- a Lua function
            table.insert(traceback, string.format("[%s]:%d", info.short_src, info.currentline))
            
            -- Gather locals
            locals[level] = {}
            local i = 1
            while true do
                local name, val = debug.getlocal(Debugger.thread.co, level + 2, i)
                if not name then break end
                locals[level][name] = { val, i }
                i = i + 1
            end
        end
        level = level + 1
    end
    
    info.traceback = table.concat(traceback, "\n")
end

-- Continue
local function continue()
    Debugger.thread = nil
    _G.c = saved_values.c
    _G.l = saved_values.l
    _G.sl = saved_values.sl
    _G.t = saved_values.t
    
    
    -- Clear traceback & locals
    info.traceback = {}
    info.locals = {}
end

-- Locals
local function locals(name, level)
    
    -- Default to display first stack frame
    name = name or 1
    
    if type(name) == "string" then
        if level == nil then
            for i,frame in ipairs(info.locals) do
                local v = frame[name]
                if v then
                    v = v[1]
                    local typ = type(v)
                    if typ == "table" or typ == "function" then
                        print(name .. " = " .. tostring(v))
                    else
                        print(name .. " = " .. typ .. ": " .. tostring(v))
                    end
                    return v
                end
            end
            
            objc.warning("No local '" .. name .. "' in any stack frame")
        else
            local frame = info.locals[level]
            if frame == nil then
                objc.warning("Invalid stack frame: " .. level)
                return
            end
            
            local v = frame[name]
            if v then
                v = v[1]
                local typ = type(v)
                if typ == "table" or typ == "function" then
                    print(name .. " = " .. tostring(v))
                else
                    print(name .. " = " .. typ .. ": " .. tostring(v))
                end
                return v
            else
                objc.warning("No local '" .. name .. "' in stack frame " .. level)
            end
        end
            
    elseif type(name) == "number" then
        if info.locals[name] == nil then
            objc.warning("Invalid stack frame: " .. name)
            return
        end
        
        local out = {"Locals (level " .. name .. "):"}
        for k,v in pairs(info.locals[name]) do
            
            v = v[1]
            local typ = type(v)
            
            if typ == "table" or typ == "function" then
                table.insert(out, k .. " = " .. tostring(v))
            else
                local vs = tostring(v)
                if string.len(vs) > 32 then
                    vs = "..."
                end
                
                table.insert(out, k .. " = " .. typ .. ": " .. vs)
            end
        end
        print(table.concat(out, "\n"))
    else
        objc.warning("l: Invalid parameter type. Expects string or number.")
    end
end

local function setlocal(name, value, level)
    
    if level == nil then
        
        for i,frame in ipairs(info.locals) do
            local v = frame[name]
            if v then
                -- Set the new value
                debug.setlocal(Debugger.thread.co, i + 2, v[2], value)
                v[1] = value
                print("Local '" .. name .. "' set in stack frame " .. i)
                return
            end
        end
            
        objc.warning("No local '" .. name .. "' in any stack frame")
    else
        
        local frame = info.locals[level]
        if frame == nil then
            objc.warning("Invalid stack frame: " .. level)
            return
        end
                
        local v = frame[name]
        if v then
            -- Set the new value
            debug.setlocal(Debugger.thread.co, level, v[2], value)
            v[1] = value
            print("Local '" .. name .. "' set in stack frame " .. level)
        else
            objc.warning("No local '" .. name .. "' in stack frame " .. level)
        end
    end
end

local function traceback()
    print("Trace:\n" .. info.traceback)
end

local function launch_debugger()
    Debugger.thread = Thread.current()
    saved_values.c = _G.c
    saved_values.l = _G.l
    saved_values.sl = _G.sl
    saved_values.t = _G.t
    _G.c = continue
    _G.l = locals
    _G.sl = setlocal
    _G.t = traceback
end

function dbg(condition)
    if condition == nil or condition == false then
        launch_debugger()
        gatherDebugInfo()
        
        -- Capture the current screen image
        local screen_image = viewer.snapshot()
        
        -- Print backtrace
        print("Debug break:\n" .. info.traceback)
        
        -- Yield until the debugger deactivates itself
        while Debugger.thread ~= nil do
            yieldframe()
            
            -- Draw the current screen so we don't flicker
            pushStyle()
            spriteMode(CORNER)
            sprite(screen_image, 0, 0, WIDTH, HEIGHT)
            popStyle()
        end
    end
end
