-- Thread management values
local threads = {}
local current_thread = nil
local sbFlushImage = image(1,1)

-- Thread management task
tween.interval(0.001, function()
    
    -- Reset per frame flags
    for _,thread in ipairs(threads) do
        thread.done_this_frame = false
    end
    
    -- Keep running while threads are still executing
    local thread_ran = true
    while thread_ran do
        
        -- Reset flag
        thread_ran = false
        
        local deadThreads = {}
    
        for i,thread in ipairs(threads) do
            
            local clock = os.clock()
            local dbg_thread = Debugger.thread
            
            -- Is the coroutine dead?
            if coroutine.status(thread.co) == "dead" then
                table.insert(deadThreads, i)
                
            -- Has the thread finished for this frame?
            elseif thread.done_this_frame then
                
            -- Ignore this thread if it's sleeping
            elseif thread.sleep_end and clock < thread.sleep_end then
                
                -- If the sleeping thread is 'blocking' pretend we ran
                if thread.blocking then
                    thread_ran = true
                end
                
            -- If we're debugging, only run the debugged thread
            elseif dbg_thread ~= nil and thread ~= dbg_thread then
                
            -- All checks have passed. Resume the thread!
            else
                -- Track the current thread
                current_thread = thread
                thread_ran = true
                
                local runMainRet = nil
                while true do
                    local res = table.pack(coroutine.resume(thread.co, runMainRet))
                    if not res[1] then
                        -- Propagate the error
                        error(res[2])
                    elseif res[2] then
                        -- Call the value returned from the resume
                        -- and resume again.
                        -- This is how a thread forces something to
                        -- run on the main thread.
                        current_thread = nil
                        runMainRet = res[2](select(3, table.unpack(res)))
                        current_thread = thread
                    else
                        -- Pause execution
                        break
                    end
                end
                
                current_thread = nil
            end
        end
        
        for _,ti in ipairs(deadThreads) do
            table.remove(threads, ti)
        end        
    end
        
    -- Render an offscreen sprite to flush the
    -- spritebatcher.
    -- This ensures sprite rendering done on
    -- threads is actually visible on screen.
    -- A wild guess is that the batcher is only
    -- flushed after the draw() call, and not
    -- after tween handling.
    spriteMode(CORNER)
    sprite(sbFlushImage, -1, -1)
end)

-- Thread functions
function yield()
    coroutine.yield()
end

function yieldframe()
    current_thread.done_this_frame = true
    coroutine.yield()
end


-- Thread class
Thread = class()
function Thread:init(fn)
    self.co = coroutine.create(fn)
    self.blocking = false
    
    table.insert(threads, self)
end

function Thread.runMain(func, ...)
    return coroutine.yield(func, ...)
end

function Thread.current()
    return current_thread
end

function Thread.allThreads()
    return threads
end

function Thread.sleep(duration)
    current_thread.sleep_end = os.clock() + duration
    yield()
end

-- These should do nothing
__wrapGlobal("setup", function() end)
__wrapGlobal("draw", function() end)

-- Thread to call setup() & draw()
Thread(function()
    
    -- We're a blocking thread
    current_thread.blocking = true
    
    -- Call setup() once
    local fn = __getWrappedGlobal("setup")
    if fn then fn() end
    
    -- Call draw() repeatedly
    while true do
        local fn = __getWrappedGlobal("draw")
        if fn then fn() end
        yieldframe()
    end
end)

-- Threaded callbacks
local callback_queue = {}
local function queue_callback(fn, ...)
    if fn == nil or Debugger.thread then return end
    table.insert(callback_queue, { 
        fn = fn,
        params = { ... }
    })
end
__wrapGlobal("touched", queue_callback)
__wrapGlobal("hover", queue_callback)
__wrapGlobal("scroll", queue_callback)
__wrapGlobal("pinch", queue_callback)
__wrapGlobal("keyboard", queue_callback)
-- __wrapGlobal("willClose", queue_callback) -- Don't wrap this

Thread(function()
    
    -- We're a blocking thread as callbacks should
    -- be handled in the same frame as they were issued
    current_thread.blocking = true
    
    while true do
        for _,cb in ipairs(callback_queue) do
            cb.fn(table.unpack(cb.params))
        end
        callback_queue = {}
        
        yieldframe()
    end
    
end)
